<?php

namespace Drupal\charts_ai_agents\Plugin\AiAgent;

use Drupal\ai_agents_extra\Plugin\AiAgent\ViewsAgent;
use Drupal\Core\Access\AccessResult;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\Core\Url;
use Drupal\ai_agents\Attribute\AiAgent;
use Drupal\ai_agents\PluginInterfaces\AiAgentInterface;
use Drupal\views\Entity\View;
use Drupal\views\Views;
use Symfony\Component\Yaml\Yaml;

/**
 * The Charts Views agent.
 */
#[AiAgent(
  id: 'charts_views_agent',
  label: new TranslatableMarkup('Charts Views Agent'),
)]
class ChartsViewsAgent extends ViewsAgent {

  /**
   * {@inheritDoc}
   */
  public function agentsNames() {
    return [
      'Charts Views Agent',
    ];
  }

  /**
   * {@inheritDoc}
   */
  public function agentsCapabilities() {
    return [
      'charts_views_agent' => [
        'name' => 'Charts Views Agent',
        'description' => "This is an agent that can create Drupal Views with a Chart format. This means that they can be used to setup a web view around an entity type, including pages, blocks and if plugins exists maps, data exports like csv, json etc. This agent only needs a valid entity type to start.",
        'usage_instructions' => "If they ask you to create a report, page or list, assume in Drupal they mean Views.",
        'inputs' => [
          'free_text' => [
            'name' => 'Prompt',
            'type' => 'string',
            'description' => 'Any information needed to generate, edit, delete or information about Drupal Views.',
            'default_value' => '',
          ],
        ],
        'outputs' => [
          'answers' => [
            'description' => 'The answers to the questions asked about the Views or the action that was taken.',
            'type' => 'string',
          ],
        ],
      ],
    ];
  }

  /**
   * {@inheritDoc}
   */
  public function isAvailable() {
    // Check if modules are installed.
    $dependencies = ['views', 'charts'];
    foreach ($dependencies as $dependency) {
      if (!$this->agentHelper->isModuleEnabled($dependency)) {
        return FALSE;
      }
    }
    return TRUE;
  }

  /**
   * {@inheritDoc}
   */
  public function isNotAvailableMessage() {
    return $this->t('You need to enable the Views and Charts modules and one of the Charts sub-modules to do this.');
  }

  /**
   * {@inheritDoc}
   */
  public function approveSolution() {
    $this->data[0]['action'] = 'create';
  }

  /**
   * {@inheritDoc}
   */
  public function hasAccess() {
    // Check for permissions.
    if (!$this->currentUser->hasPermission('administer views')) {
      return AccessResult::forbidden();
    }
    return parent::hasAccess();
  }

  /**
   * {@inheritDoc}
   */
  public function determineSolvability() {
    $this->determineSolvabilityCharts();
    $this->data = $this->determineTypeOfTask();

    return match ($this->data[0]['action']) {
      'create', 'edit', 'delete' => AiAgentInterface::JOB_SOLVABLE,
      'information' => AiAgentInterface::JOB_SHOULD_ANSWER_QUESTION,
      default => AiAgentInterface::JOB_NOT_SOLVABLE,
    };

  }

  /**
   * {@inheritDoc}
   */
  public function determineSolvabilityCharts() {
    // Setup the helper runner.
    $this->agentHelper->setupRunner($this);
    return AiAgentInterface::JOB_SOLVABLE;
  }

  /**
   * {@inheritDoc}
   */
  public function askQuestion() {
    return $this->questions;
  }

  /**
   * {@inheritDoc}
   */
  public function solve() {
    // Get the view executable.
    $this->getViewExecCharts();
    $message = '';
    switch ($this->data[0]['action']) {
      case 'create':
        try {
          // We need to build this in steps.
          if (empty($this->data[0]['show_mode']) || $this->data[0]['show_mode'] == 'content') {
            $this->determineViewType();
          }
          else {
            $this->determineFieldTypesCharts();
          }
          $this->determineFilterTypes();
          $message = "Your view is here: " . Url::fromRoute('entity.view.edit_form', [
            'view' => $this->data[0]['data_name'],
          ])->toString();
        }
        catch (\Exception $e) {
          $message = 'There was an error creating the node type: ' . $e->getMessage();
        }
        break;

      case 'edit':
        // @todo Implement edit.
        break;

      case 'delete':
        $message = 'This agent does not support deleting views.';
        break;

      case 'fail':
        return $this->data[0]['extra_info'];

      case 'information':
        return $this->answerQuestion();
    }
    return $message;
  }

  /**
   * Determine if the context is asking a question or wants a audit done.
   *
   * @return string
   *   The context.
   *
   * @throws \Exception
   */
  public function determineTypeOfTask() {
    $data = $this->agentHelper->runSubAgent('determineViewsTaskCharts', [
      'Possible Views Types' => $this->getViewsTypes(),
      'Possible Views Styles' => $this->getViewsStyles(),
      'Entity Types and Bundles' => $this->getEntityTypesAndBundles(),
    ]);

    if (isset($data[0]['action']) && in_array($data[0]['action'], [
      'create',
      'edit',
      'delete',
      'information',
      'fail',
    ])) {
      return $data;
    }
    throw new \Exception('Invalid action in determining the task.');
  }

  /**
   * Gets the views executable.
   */
  protected function getViewExecCharts() {
    try {
      $this->viewExec = Views::getView($this->data[0]['data_name']);
      if (!empty($this->viewExec)) {
        return;
      }
    }
    catch (\Exception $e) {
      // Do nothing.
    }

    // If its create, we create a new view.
    if ($this->data[0]['action'] === 'create') {
      $display = [
        'id' => 'default',
        'display_title' => 'default',
        'enabled' => TRUE,
        'display_options' => [
          'title' => $this->data[0]['title'],
          'use_pager' => $this->data[0]['pager'] == 'none' ? FALSE : TRUE,
          'style' => [
            'type' => $this->data[0]['view_style'],
          ],
          'pager' => [
            'type' => $this->data[0]['pager'],
            'options' => [
              'items_per_page' => $this->data[0]['amount_per_page'],
            ],
          ],
          'path' => trim($this->data[0]['path'], '/'),
        ],
        'display_plugin' => 'default',
        'position' => 0,
        'cache_metadata' => [
          'tags' => [],
        ],
      ];
      if ($this->data[0]['view_style'] == 'chart') {
        $display['display_options']['style']['options'] = [
          'chart_settings' => $this->data[0]['chart_settings'],
        ];
      }
      if (isset($this->data[0]['permission'])) {
        $display['display_options']['access']['type'] = 'perm';
        $display['display_options']['access']['options']['perm'] = $this->data[0]['permission'];
      }
      // Create another display as needed.
      $type = $this->data[0]['view_type'];
      $visibleDisplay = [
        'id' => $type . '_1',
        'display_title' => $this->data[0]['title'],
        'display_plugin' => $type,
        'enabled' => TRUE,
        'position' => 1,
        'display_options' => [
          'path' => trim($this->data[0]['path'], '/'),
          'display_extenders' => [],
        ],
        'cache_metadata' => [
          'tags' => [],
          'max-age' => 0,
        ],
      ];
      $view = View::create([
        'id' => $this->data[0]['data_name'],
        'label' => $this->data[0]['title'],
        'module' => 'views',
        'description' => $this->data[0]['description'],
        'tag' => 'default',
        // @todo load from config.
        'base_table' => $this->data[0]['entity_type'] . '_field_data',
        'display' => [
          'default' => $display,
          $type . '_1' => $visibleDisplay,
        ],
      ]);

      try {
        $view->save();
      }
      catch (\Exception $e) {
        throw new \Exception('Could not save the view: ' . $e->getMessage());
      }
      $this->garbage[] = $view;
      $this->viewExec = Views::getView($this->data[0]['data_name']);
    }
    else {
      $this->viewExec = Views::getView($this->data[0]['data_name']);
    }
    if (empty($this->viewExec)) {
      throw new \Exception('The view does not exist.');
    }
  }

  /**
   * Determine the field types.
   */
  public function determineFieldTypesCharts() {
    $data = $this->agentHelper->runSubAgent('determineFieldsTaskCharts', [
      'Requested Field Types' => !empty($this->data[0]['field_types']) ? implode(', ', $this->data[0]['field_types']) : 'No field types specified.',
      'Available Field Types' => $this->getFieldTypes(),
      'Available Plugin Types' => $this->getFieldPluginOptions(),
    ]);
    if (empty($data[0]['action'])) {
      throw new \Exception('We could not determine which field types was needed for the view.');
    }
    // Load the display, to add fields.
    $display = $this->viewExec->getDisplay('default');
    $fieldsDisplay = $display->getOption('fields');

    // Then try to generate settings per field.
    foreach ($data as $field) {
      // Check where this field is stored.
      $table = $this->getStorageTable($field['field'], $this->data[0]['entity_type']);
      if ($table || $field['plugin_id'] == 'entity_operations') {
        // Get the config for the field type.
        $options = $this->getFieldConfigs($field['plugin_id']);
        // Set some values directly.
        $setOptions = [
          'id' => $field['id'],
          'table' => $table,
          'field' => $field['field'],
          'label' => $field['label'],
          'admin_label' => $field['label'],
          'plugin_id' => $field['plugin_id'],
          'exclude' => $field['exclude'],
        ];
        $fieldData = $this->agentHelper->runSubAgent('determineFieldConfig', [
          'Field Type' => $field['type'] ?? 'No type specified.',
          'View Style' => $this->data[0]['view_style'],
          'Field Configuration' => Yaml::dump($options['mapping'], 12, 4),
          'Current Options' => Yaml::dump($setOptions, 12, 4),
        ]);

        if ($fieldData[0]['action'] == 'change') {
          $setOptions = array_merge_recursive($setOptions, $fieldData[0]['new_config']);
        }
        // Some special cases.
        if (!empty($field['type'])) {
          switch ($field['type']) {
            case 'image':
              if ($this->data[0]['view_style'] == 'table') {
                $setOptions['settings'] = [
                  'image_style' => 'thumbnail',
                  'image_link' => 'content',
                ];
              }
              break;
          }
        }
        if ($field['plugin_id'] == 'entity_operations') {
          $setOptions['table'] = 'node';
        }
        // Generate a views field.
        $fieldsDisplay[$setOptions['id']] = $setOptions;
      }
    }
    $display->setOption('fields', $fieldsDisplay);
    $this->viewExec->setDisplay('default', $display);
    $this->viewExec->save();
  }

}
